from datetime import date

import ibis
import pytest
from packaging import version
from pytest_lazyfixture import lazy_fixture
from substrait_validator import Config, check_plan

from ibis_substrait.compiler.decompile import decompile


@pytest.fixture
def tpc_h01(lineitem):
    return (
        lineitem.filter(lambda t: t.l_shipdate <= date(year=1998, month=9, day=2))
        .group_by(["l_returnflag", "l_linestatus"])
        .aggregate(
            sum_qty=lambda t: t.l_quantity.sum(),
            sum_base_price=lambda t: t.l_extendedprice.sum(),
            sum_disc_price=lambda t: (t.l_extendedprice * (1 - t.l_discount)).sum(),
            sum_charge=lambda t: (
                t.l_extendedprice * (1 - t.l_discount) * (1 + t.l_tax)
            ).sum(),
            avg_qty=lambda t: t.l_quantity.mean(),
            avg_price=lambda t: t.l_extendedprice.mean(),
            avg_disc=lambda t: t.l_discount.mean(),
            count_order=lambda t: t.count(),
        )
        .sort_by(["l_returnflag", "l_linestatus"])
    )


@pytest.fixture
def tpc_h02(
    part, supplier, partsupp, nation, region, REGION="EUROPE", SIZE=25, TYPE="BRASS"
):
    "Minimum Cost Supplier Query (Q2)"

    expr = (
        part.join(partsupp, part.p_partkey == partsupp.ps_partkey)
        .join(supplier, supplier.s_suppkey == partsupp.ps_suppkey)
        .join(nation, supplier.s_nationkey == nation.n_nationkey)
        .join(region, nation.n_regionkey == region.r_regionkey)
    )

    subexpr = (
        partsupp.join(supplier, supplier.s_suppkey == partsupp.ps_suppkey)
        .join(nation, supplier.s_nationkey == nation.n_nationkey)
        .join(region, nation.n_regionkey == region.r_regionkey)
    )

    subexpr = subexpr[
        (subexpr.r_name == REGION) & (expr.p_partkey == subexpr.ps_partkey)
    ]

    filters = [
        expr.p_size == SIZE,
        expr.p_type.like("%" + TYPE),
        expr.r_name == REGION,
        expr.ps_supplycost == subexpr.ps_supplycost.min(),
    ]
    q = expr.filter(filters)

    q = q.select(
        [
            q.s_acctbal,
            q.s_name,
            q.n_name,
            q.p_partkey,
            q.p_mfgr,
            q.s_address,
            q.s_phone,
            q.s_comment,
        ]
    )

    return q.sort_by(
        [
            ibis.desc(q.s_acctbal),
            q.n_name,
            q.s_name,
            q.p_partkey,
        ]
    ).limit(100)


@pytest.fixture
def tpc_h03(
    customer,
    orders,
    lineitem,
):
    DATE = "1995-03-15"
    q = customer.join(orders, customer.c_custkey == orders.o_custkey)
    q = q.join(lineitem, lineitem.l_orderkey == orders.o_orderkey)
    q = q.filter(
        [q.c_mktsegment == "BUILDING", q.o_orderdate < DATE, q.l_shipdate > DATE]
    )
    qg = q.group_by([q.l_orderkey, q.o_orderdate, q.o_shippriority])
    q = qg.aggregate(revenue=(q.l_extendedprice * (1 - q.l_discount)).sum())
    q = q.sort_by([ibis.desc(q.revenue), q.o_orderdate])
    q = q.limit(10)

    return q


@pytest.fixture
def tpc_h04(orders, lineitem):
    cond = (lineitem.l_orderkey == orders.o_orderkey) & (
        lineitem.l_commitdate < lineitem.l_receiptdate
    )
    q = orders.filter(
        [
            cond.any(),
            orders.o_orderdate >= "1993-07-01",
            orders.o_orderdate < "1993-10-01",
        ]
    )
    q = q.group_by([orders.o_orderpriority])
    q = q.aggregate(order_count=orders.count())
    q = q.sort_by([orders.o_orderpriority])
    return q


@pytest.fixture
def tpc_h05(customer, orders, lineitem, supplier, nation, region):
    q = customer
    q = q.join(orders, customer.c_custkey == orders.o_custkey)
    q = q.join(lineitem, lineitem.l_orderkey == orders.o_orderkey)
    q = q.join(supplier, lineitem.l_suppkey == supplier.s_suppkey)
    q = q.join(
        nation,
        (customer.c_nationkey == supplier.s_nationkey)
        & (supplier.s_nationkey == nation.n_nationkey),
    )
    q = q.join(region, nation.n_regionkey == region.r_regionkey)

    q = q.filter(
        [
            q.r_name == "ASIA",
            q.o_orderdate >= "1994-01-01",
            q.o_orderdate < "1995-01-01",
        ]
    )
    revexpr = q.l_extendedprice * (1 - q.l_discount)
    gq = q.group_by([q.n_name])
    q = gq.aggregate(revenue=revexpr.sum())
    q = q.sort_by([ibis.desc(q.revenue)])
    return q


@pytest.fixture
def tpc_h06(lineitem):
    q = lineitem
    discount_min = round(0.06 - 0.01, 2)
    discount_max = round(0.06 + 0.01, 2)
    q = q.filter(
        [
            q.l_shipdate >= "1994-01-01",
            q.l_shipdate < "1995-01-01",
            q.l_discount.between(discount_min, discount_max),
            q.l_quantity < 24,
        ]
    )
    q = q.aggregate(revenue=(q.l_extendedprice * q.l_discount).sum())
    return q


@pytest.fixture
def tpc_h07(supplier, lineitem, orders, customer, nation):
    q = supplier
    q = q.join(lineitem, supplier.s_suppkey == lineitem.l_suppkey)
    q = q.join(orders, orders.o_orderkey == lineitem.l_orderkey)
    q = q.join(customer, customer.c_custkey == orders.o_custkey)
    n1 = nation
    n2 = nation.view()
    q = q.join(n1, supplier.s_nationkey == n1.n_nationkey)
    q = q.join(n2, customer.c_nationkey == n2.n_nationkey)

    q = q[
        n1.n_name.name("supp_nation"),
        n2.n_name.name("cust_nation"),
        lineitem.l_shipdate,
        lineitem.l_extendedprice,
        lineitem.l_discount,
        lineitem.l_shipdate.year().cast("string").name("l_year"),
        (lineitem.l_extendedprice * (1 - lineitem.l_discount)).name("volume"),
    ]

    q = q.filter(
        [
            ((q.cust_nation == "FRANCE") & (q.supp_nation == "GERMANY"))
            | ((q.cust_nation == "GERMANY") & (q.supp_nation == "FRANCE")),
            q.l_shipdate.between("1995-01-01", "1996-12-31"),
        ]
    )

    gq = q.group_by(["supp_nation", "cust_nation", "l_year"])
    q = gq.aggregate(revenue=q.volume.sum())
    q = q.sort_by(["supp_nation", "cust_nation", "l_year"])

    return q


@pytest.fixture
def tpc_h08(
    part,
    supplier,
    lineitem,
    orders,
    customer,
    region,
    nation,
):
    n1 = nation
    n2 = n1.view()

    q = part
    q = q.join(lineitem, part.p_partkey == lineitem.l_partkey)
    q = q.join(supplier, supplier.s_suppkey == lineitem.l_suppkey)
    q = q.join(orders, lineitem.l_orderkey == orders.o_orderkey)
    q = q.join(customer, orders.o_custkey == customer.c_custkey)
    q = q.join(n1, customer.c_nationkey == n1.n_nationkey)
    q = q.join(region, n1.n_regionkey == region.r_regionkey)
    q = q.join(n2, supplier.s_nationkey == n2.n_nationkey)

    q = q[
        orders.o_orderdate.year().cast("string").name("o_year"),
        (lineitem.l_extendedprice * (1 - lineitem.l_discount)).name("volume"),
        n2.n_name.name("nation"),
        region.r_name,
        orders.o_orderdate,
        part.p_type,
    ]

    q = q.filter(
        [
            q.r_name == "AMERICA",
            q.o_orderdate.between("1995-01-01", "1996-12-31"),
            q.p_type == "ECONOMY ANODIZED STEEL",
        ]
    )

    q = q.mutate(
        nation_volume=ibis.case().when(q.nation == "BRAZIL", q.volume).else_(0).end()
    )
    gq = q.group_by([q.o_year])
    q = gq.aggregate(mkt_share=q.nation_volume.sum() / q.volume.sum())
    q = q.sort_by([q.o_year])
    return q


@pytest.fixture
def tpc_h09(part, supplier, lineitem, partsupp, orders, nation):

    q = lineitem
    q = q.join(supplier, supplier.s_suppkey == lineitem.l_suppkey)
    q = q.join(
        partsupp,
        (partsupp.ps_suppkey == lineitem.l_suppkey)
        & (partsupp.ps_partkey == lineitem.l_partkey),
    )
    q = q.join(part, part.p_partkey == lineitem.l_partkey)
    q = q.join(orders, orders.o_orderkey == lineitem.l_orderkey)
    q = q.join(nation, supplier.s_nationkey == nation.n_nationkey)

    q = q[
        (q.l_extendedprice * (1 - q.l_discount) - q.ps_supplycost * q.l_quantity).name(
            "amount"
        ),
        q.o_orderdate.year().cast("string").name("o_year"),
        q.n_name.name("nation"),
        q.p_name,
    ]

    q = q.filter([q.p_name.like("%GREEN%")])

    gq = q.group_by([q.nation, q.o_year])
    q = gq.aggregate(sum_profit=q.amount.sum())
    q = q.sort_by([q.nation, ibis.desc(q.o_year)])
    return q


@pytest.fixture
def tpc_h10(customer, orders, lineitem, nation):
    q = customer
    q = q.join(orders, customer.c_custkey == orders.o_custkey)
    q = q.join(lineitem, lineitem.l_orderkey == orders.o_orderkey)
    q = q.join(nation, customer.c_nationkey == nation.n_nationkey)

    q = q.filter(
        [
            (q.o_orderdate >= "1993-01-01") & (q.o_orderdate < "1993-04-01"),
            q.l_returnflag == "R",
        ]
    )

    gq = q.group_by(
        [
            q.c_custkey,
            q.c_name,
            q.c_acctbal,
            q.c_phone,
            q.n_name,
            q.c_address,
            q.c_comment,
        ]
    )
    q = gq.aggregate(revenue=(q.l_extendedprice * (1 - q.l_discount)).sum())

    q = q.sort_by(ibis.desc(q.revenue))
    return q.limit(20)


@pytest.fixture
def tpc_h11(partsupp, supplier, nation):
    q = partsupp
    q = q.join(supplier, partsupp.ps_suppkey == supplier.s_suppkey)
    q = q.join(nation, nation.n_nationkey == supplier.s_nationkey)

    q = q.filter([q.n_name == "GERMANY"])

    innerq = partsupp
    innerq = innerq.join(supplier, partsupp.ps_suppkey == supplier.s_suppkey)
    innerq = innerq.join(nation, nation.n_nationkey == supplier.s_nationkey)
    innerq = innerq.filter([innerq.n_name == "GERMANY"])
    innerq = innerq.aggregate(total=(innerq.ps_supplycost * innerq.ps_availqty).sum())

    gq = q.group_by([q.ps_partkey])
    q = gq.aggregate(value=(q.ps_supplycost * q.ps_availqty).sum())
    q = q.filter([q.value > innerq.total * 0.0001])
    q = q.sort_by(ibis.desc(q.value))
    return q


@pytest.fixture
def tpc_h12(orders, lineitem):
    q = orders
    q = q.join(lineitem, orders.o_orderkey == lineitem.l_orderkey)

    q = q.filter(
        [
            q.l_shipmode.isin(["MAIL", "SHIP"]),
            q.l_commitdate < q.l_receiptdate,
            q.l_shipdate < q.l_commitdate,
            q.l_receiptdate >= "1994-01-01",
            q.l_receiptdate < "1995-01-01",
        ]
    )

    gq = q.group_by([q.l_shipmode])
    q = gq.aggregate(
        high_line_count=(
            q.o_orderpriority.case()
            .when("1-URGENT", 1)
            .when("2-HIGH", 1)
            .else_(0)
            .end()
        ).sum(),
        low_line_count=(
            q.o_orderpriority.case()
            .when("1-URGENT", 0)
            .when("2-HIGH", 0)
            .else_(1)
            .end()
        ).sum(),
    )
    q = q.sort_by(q.l_shipmode)

    return q


@pytest.fixture
def tpc_h13(customer, orders):
    innerq = customer
    innerq = innerq.left_join(
        orders,
        (customer.c_custkey == orders.o_custkey)
        & ~orders.o_comment.like("%special%requests%"),
    )
    innergq = innerq.group_by([innerq.c_custkey])
    innerq = innergq.aggregate(c_count=innerq.o_orderkey.count())

    gq = innerq.group_by([innerq.c_count])
    q = gq.aggregate(custdist=innerq.count())

    q = q.sort_by([ibis.desc(q.custdist), ibis.desc(q.c_count)])
    return q


@pytest.fixture
def tpc_h14(lineitem, part):
    q = lineitem
    q = q.join(part, lineitem.l_partkey == part.p_partkey)
    q = q.filter([q.l_shipdate >= "1995-09-01", q.l_shipdate < "1995-10-01"])

    revenue = q.l_extendedprice * (1 - q.l_discount)
    promo_revenue = q.p_type.like("PROMO%").ifelse(revenue, 0)

    q = q.aggregate(promo_revenue=100 * promo_revenue.sum() / revenue.sum())
    return q


@pytest.fixture
def tpc_h15(lineitem, supplier):
    qrev = lineitem
    qrev = qrev.filter(
        [lineitem.l_shipdate >= "1996-01-01", lineitem.l_shipdate < "1996-04-01"]
    )

    gqrev = qrev.group_by([lineitem.l_suppkey])
    qrev = gqrev.aggregate(
        total_revenue=(qrev.l_extendedprice * (1 - qrev.l_discount)).sum()
    )

    q = supplier.join(qrev, supplier.s_suppkey == qrev.l_suppkey)
    q = q.filter([q.total_revenue == qrev.total_revenue.max()])
    q = q.sort_by([q.s_suppkey])
    q = q[q.s_suppkey, q.s_name, q.s_address, q.s_phone, q.total_revenue]
    return q


@pytest.fixture
def tpc_h16(partsupp, part, supplier):
    q = partsupp.join(part, part.p_partkey == partsupp.ps_partkey)
    q = q.filter(
        [
            q.p_brand != "Brand#45",
            ~q.p_type.like("MEDIUM POLISHED%"),
            q.p_size.isin((49, 14, 23, 45, 19, 3, 36, 9)),
            ~q.ps_suppkey.isin(
                supplier.filter(
                    [supplier.s_comment.like("%Customer%Complaints%")]
                ).s_suppkey
            ),
        ]
    )
    gq = q.groupby([q.p_brand, q.p_type, q.p_size])
    q = gq.aggregate(supplier_cnt=q.ps_suppkey.nunique())
    q = q.sort_by([ibis.desc(q.supplier_cnt), q.p_brand, q.p_type, q.p_size])
    return q


@pytest.fixture
def tpc_h17(lineitem, part):
    q = lineitem.join(part, part.p_partkey == lineitem.l_partkey)

    innerq = lineitem
    innerq = innerq.filter([innerq.l_partkey == q.p_partkey])

    q = q.filter(
        [
            q.p_brand == "Brand#23",
            q.p_container == "MED BOX",
            q.l_quantity < (0.2 * innerq.l_quantity.mean()),
        ]
    )
    q = q.aggregate(avg_yearly=q.l_extendedprice.sum() / 7.0)
    return q


@pytest.fixture
def tpc_h18(customer, orders, lineitem):
    subgq = lineitem.groupby([lineitem.l_orderkey])
    subq = subgq.aggregate(qty_sum=lineitem.l_quantity.sum())
    subq = subq.filter([subq.qty_sum > 300])

    q = customer
    q = q.join(orders, customer.c_custkey == orders.o_custkey)
    q = q.join(lineitem, orders.o_orderkey == lineitem.l_orderkey)
    q = q.filter([q.o_orderkey.isin(subq.l_orderkey)])

    gq = q.groupby([q.c_name, q.c_custkey, q.o_orderkey, q.o_orderdate, q.o_totalprice])
    q = gq.aggregate(sum_qty=q.l_quantity.sum())
    q = q.sort_by([ibis.desc(q.o_totalprice), q.o_orderdate])
    return q.limit(100)


@pytest.fixture
def tpc_h19(lineitem, part):
    q = lineitem.join(part, part.p_partkey == lineitem.l_partkey)

    q1 = (
        (q.p_brand == "Brand#12")
        & (q.p_container.isin(("SM CASE", "SM BOX", "SM PACK", "SM PKG")))
        & (q.l_quantity >= 1)
        & (q.l_quantity <= 1 + 10)
        & (q.p_size.between(1, 5))
        & (q.l_shipmode.isin(("AIR", "AIR REG")))
        & (q.l_shipinstruct == "DELIVER IN PERSON")
    )

    q2 = (
        (q.p_brand == "Brand#23")
        & (q.p_container.isin(("MED BAG", "MED BOX", "MED PKG", "MED PACK")))
        & (q.l_quantity >= 10)
        & (q.l_quantity <= 10 + 10)
        & (q.p_size.between(1, 10))
        & (q.l_shipmode.isin(("AIR", "AIR REG")))
        & (q.l_shipinstruct == "DELIVER IN PERSON")
    )

    q3 = (
        (q.p_brand == "Brand#34")
        & (q.p_container.isin(("LG CASE", "LG BOX", "LG PACK", "LG PKG")))
        & (q.l_quantity >= 20)
        & (q.l_quantity <= 20 + 10)
        & (q.p_size.between(1, 15))
        & (q.l_shipmode.isin(("AIR", "AIR REG")))
        & (q.l_shipinstruct == "DELIVER IN PERSON")
    )

    q = q.filter([q1 | q2 | q3])
    q = q.aggregate(revenue=(q.l_extendedprice * (1 - q.l_discount)).sum())
    return q


@pytest.fixture
def tpc_h20(supplier, nation, partsupp, part, lineitem):
    q1 = supplier.join(nation, supplier.s_nationkey == nation.n_nationkey)

    q3 = part.filter([part.p_name.like("forest%")])
    q2 = partsupp

    q4 = lineitem.filter(
        [
            lineitem.l_partkey == q2.ps_partkey,
            lineitem.l_suppkey == q2.ps_suppkey,
            lineitem.l_shipdate >= "1994-01-01",
            lineitem.l_shipdate < "1995-01-01",
        ]
    )

    q2 = q2.filter(
        [
            partsupp.ps_partkey.isin(q3.p_partkey),
            partsupp.ps_availqty > 0.5 * q4.l_quantity.sum(),
        ]
    )

    q1 = q1.filter([q1.n_name == "CANADA", q1.s_suppkey.isin(q2.ps_suppkey)])

    q1 = q1[q1.s_name, q1.s_address]

    return q1.sort_by(q1.s_name)


@pytest.fixture
def tpc_h21(supplier, lineitem, orders, nation):
    L2 = lineitem.view()
    L3 = lineitem.view()

    q = supplier
    q = q.join(lineitem, supplier.s_suppkey == lineitem.l_suppkey)
    q = q.join(orders, orders.o_orderkey == lineitem.l_orderkey)
    q = q.join(nation, supplier.s_nationkey == nation.n_nationkey)
    q = q[
        q.l_orderkey.name("l1_orderkey"),
        q.o_orderstatus,
        q.l_receiptdate,
        q.l_commitdate,
        q.l_suppkey.name("l1_suppkey"),
        q.s_name,
        q.n_name,
    ]
    q = q.filter(
        [
            q.o_orderstatus == "F",
            q.l_receiptdate > q.l_commitdate,
            q.n_name == "SAUDI ARABIA",
            ((L2.l_orderkey == q.l1_orderkey) & (L2.l_suppkey != q.l1_suppkey)).any(),
            ~(
                (
                    (L3.l_orderkey == q.l1_orderkey)
                    & (L3.l_suppkey != q.l1_suppkey)
                    & (L3.l_receiptdate > L3.l_commitdate)
                ).any()
            ),
        ]
    )

    gq = q.group_by([q.s_name])
    q = gq.aggregate(numwait=q.count())
    q = q.sort_by([ibis.desc(q.numwait), q.s_name])
    return q.limit(100)


@pytest.fixture
def tpc_h22(customer, orders):
    q = customer.filter(
        [
            customer.c_acctbal > 0.00,
            customer.c_phone.substr(0, 2).isin(
                ("13", "31", "23", "29", "30", "18", "17")
            ),
        ]
    )
    q = q.aggregate(avg_bal=customer.c_acctbal.mean())

    custsale = customer.filter(
        [
            customer.c_phone.substr(0, 2).isin(
                ("13", "31", "23", "29", "30", "18", "17")
            ),
            customer.c_acctbal > q.avg_bal,
            ~(orders.o_custkey == customer.c_custkey).any(),
        ]
    )
    custsale = custsale[
        customer.c_phone.substr(0, 2).name("cntrycode"), customer.c_acctbal
    ]

    gq = custsale.group_by(custsale.cntrycode)
    outerq = gq.aggregate(numcust=custsale.count(), totacctbal=custsale.c_acctbal.sum())

    return outerq.sort_by(outerq.cntrycode)


@pytest.mark.xfail(
    version.parse("2.1.1") < version.parse(ibis.__version__) <= version.parse("3.0.2"),
    reason="issue with unbounded decimal precision fixed on ibis-master",
)
def test_tpch1(tpc_h01, lineitem, compiler):
    plan = compiler.compile(tpc_h01)
    assert plan.SerializeToString()

    (result,) = decompile(plan)
    expected = (
        lineitem.filter(lambda t: t.l_shipdate <= date(year=1998, month=9, day=2))
        .group_by(["l_returnflag", "l_linestatus"])
        .aggregate(
            sum_qty=lambda t: t.l_quantity.sum(),
            sum_base_price=lambda t: t.l_extendedprice.sum(),
            sum_disc_price=lambda t: (t.l_extendedprice * (1 - t.l_discount)).sum(),
            sum_charge=lambda t: (
                t.l_extendedprice * (1 - t.l_discount) * (1 + t.l_tax)
            ).sum(),
            avg_qty=lambda t: t.l_quantity.mean(),
            avg_price=lambda t: t.l_extendedprice.mean(),
            avg_disc=lambda t: t.l_discount.mean(),
            count_order=lambda t: t.count(),
        )
        .sort_by(["l_returnflag", "l_linestatus"])
    )
    assert result.equals(expected)


TPC_H = [
    lazy_fixture("tpc_h01"),
    pytest.param(
        lazy_fixture("tpc_h02"),
        marks=pytest.mark.xfail(
            raises=KeyError, reason="TableArrayView not implemented"
        ),
    ),
    lazy_fixture("tpc_h03"),
    pytest.param(
        lazy_fixture("tpc_h04"),
        marks=pytest.mark.xfail(
            raises=KeyError, reason="ExistsSubquery not implemented"
        ),
    ),
    lazy_fixture("tpc_h05"),
    lazy_fixture("tpc_h06"),
    pytest.param(
        lazy_fixture("tpc_h07"),
        marks=pytest.mark.xfail(
            raises=NotImplementedError, reason="Self reference (view)"
        ),
    ),
    pytest.param(
        lazy_fixture("tpc_h08"),
        marks=pytest.mark.xfail(
            raises=NotImplementedError, reason="Self reference (view)"
        ),
    ),
    lazy_fixture("tpc_h09"),
    lazy_fixture("tpc_h10"),
    lazy_fixture("tpc_h11"),
    lazy_fixture("tpc_h12"),
    lazy_fixture("tpc_h13"),
    pytest.param(
        lazy_fixture("tpc_h14"),
        marks=pytest.mark.xfail(
            raises=TypeError,
            reason="protobuf error resulting subquery (cannot merge Expression and AggregateFunction)",  # noqa
        ),
    ),
    pytest.param(
        lazy_fixture("tpc_h15"),
        marks=pytest.mark.xfail(
            raises=KeyError, reason="TableArrayView not implemented"
        ),
    ),
    pytest.param(
        lazy_fixture("tpc_h16"),
        marks=pytest.mark.xfail(
            raises=TypeError, reason="IntegerColumn is not iterable"
        ),
    ),
    pytest.param(
        lazy_fixture("tpc_h17"),
        marks=pytest.mark.xfail(
            raises=NotImplementedError,
            reason="ibis.expr.operations.relations.Aggregation",
        ),
    ),
    pytest.param(
        lazy_fixture("tpc_h18"),
        marks=pytest.mark.xfail(
            raises=TypeError, reason="IntegerColumn is not iterable"
        ),
    ),
    lazy_fixture("tpc_h19"),
    pytest.param(
        lazy_fixture("tpc_h20"),
        marks=pytest.mark.xfail(
            raises=TypeError, reason="IntegerColumn is not iterable"
        ),
    ),
    pytest.param(
        lazy_fixture("tpc_h21"),
        marks=pytest.mark.xfail(
            raises=KeyError, reason="ExistsSubquery not implemented"
        ),
    ),
    pytest.param(
        lazy_fixture("tpc_h22"),
        marks=pytest.mark.xfail(
            raises=KeyError, reason="NotExistsSubquery not implemented"
        ),
    ),
]


@pytest.mark.skipif(
    version.parse(ibis.__version__) <= version.parse("3.0.0"),
    reason="TPC queries aren't formatted in an ibis 2.x compatible way",
)
@pytest.mark.parametrize(
    "query",
    TPC_H,
)
def test_compile(query, compiler):
    _ = compiler.compile(query)


@pytest.mark.skipif(
    version.parse(ibis.__version__) <= version.parse("3.0.0"),
    reason="TPC queries aren't formatted in an ibis 2.x compatible way",
)
@pytest.mark.parametrize(
    "query",
    TPC_H,
)
def test_compile_validate(query, compiler):
    plan = compiler.compile(query)

    c = Config()
    # too few field names
    c.override_diagnostic_level(4003, "error", "info")
    # function def unavailable, cannot check validity of call
    c.override_diagnostic_level(6003, "warning", "info")
    # failed to resolve YAML: unknown url type
    c.override_diagnostic_level(2002, "warning", "info")  # too few field names
    # typecast validation rules are net yet implemented
    c.override_diagnostic_level(1, "warning", "info")  # too few field names

    # check_plan takes substrait Plans as arguments but because of namespacing,
    # our plan is no longer strictly a substrait.plan_pb2.Plan.
    # Workaround for now is to send in the bytes which it handles without issue.
    assert check_plan(plan.SerializeToString(), config=c)



import ibis.expr.operations as ops

IBIS_SUBSTRAIT_OP_MAPPING = {
    "Add": "add",
    "And": "and",
    "Any": "any",
    "Between": "between",
    "Count": "count",
    "CountDistinct": "countdistinct",
    "Divide": "divide",
    "Equals": "equal",
    "ExtractYear": "extract",
    "ExtractMonth": "extract",
    "ExtractDay": "extract",
    "Greater": "gt",
    "GreaterEqual": "gte",
    "Less": "lt",
    "LessEqual": "lte",
    "Max": "max",
    "Mean": "mean",
    "Min": "min",
    "Multiply": "multiply",
    "Not": "not",
    "NotEquals": "not_equal",
    "Or": "or",
    "StringLength": "string_length",  # don't know
    "StringSQLLike": "like",
    "Substring": "substring",
    "Subtract": "subtract",
    "Sum": "sum",
}

SUBSTRAIT_IBIS_OP_MAPPING = {
    v: getattr(ops, k) for k, v in IBIS_SUBSTRAIT_OP_MAPPING.items()
}
# override when reversing many-to-one mappings
SUBSTRAIT_IBIS_OP_MAPPING["extract"] = lambda span, table: getattr(
    ops, f"Extract{span.capitalize()}"
)(table)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
from ... import substrait
import typing
import typing_extensions
DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class FunctionSignature(google.protobuf.message.Message):
    """List of function signatures available."""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class FinalArgVariadic(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        class _ParameterConsistency:
            ValueType = typing.NewType('ValueType', builtins.int)
            V: typing_extensions.TypeAlias = ValueType

        class _ParameterConsistencyEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[FunctionSignature.FinalArgVariadic._ParameterConsistency.ValueType], builtins.type):
            DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
            PARAMETER_CONSISTENCY_UNSPECIFIED: FunctionSignature.FinalArgVariadic._ParameterConsistency.ValueType
            PARAMETER_CONSISTENCY_CONSISTENT: FunctionSignature.FinalArgVariadic._ParameterConsistency.ValueType
            'All argument must be the same concrete type.'
            PARAMETER_CONSISTENCY_INCONSISTENT: FunctionSignature.FinalArgVariadic._ParameterConsistency.ValueType
            'Each argument can be any possible concrete type afforded by the bounds\n            of any parameter defined in the arguments specification.\n            '

        class ParameterConsistency(_ParameterConsistency, metaclass=_ParameterConsistencyEnumTypeWrapper):
            pass
        PARAMETER_CONSISTENCY_UNSPECIFIED: FunctionSignature.FinalArgVariadic.ParameterConsistency.ValueType
        PARAMETER_CONSISTENCY_CONSISTENT: FunctionSignature.FinalArgVariadic.ParameterConsistency.ValueType
        'All argument must be the same concrete type.'
        PARAMETER_CONSISTENCY_INCONSISTENT: FunctionSignature.FinalArgVariadic.ParameterConsistency.ValueType
        'Each argument can be any possible concrete type afforded by the bounds\n        of any parameter defined in the arguments specification.\n        '
        MIN_ARGS_FIELD_NUMBER: builtins.int
        MAX_ARGS_FIELD_NUMBER: builtins.int
        CONSISTENCY_FIELD_NUMBER: builtins.int
        min_args: builtins.int
        'the minimum number of arguments allowed for the list of final arguments\n        (inclusive).\n        '
        max_args: builtins.int
        'the maximum number of arguments allowed for the list of final arguments\n        (exclusive)\n        '
        consistency: global___FunctionSignature.FinalArgVariadic.ParameterConsistency.ValueType
        'the type of parameterized type consistency'

        def __init__(self, *, min_args: builtins.int=..., max_args: builtins.int=..., consistency: global___FunctionSignature.FinalArgVariadic.ParameterConsistency.ValueType=...) -> None:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['consistency', b'consistency', 'max_args', b'max_args', 'min_args', b'min_args']) -> None:
            ...

    class FinalArgNormal(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        def __init__(self) -> None:
            ...

    class Scalar(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        ARGUMENTS_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        DESCRIPTION_FIELD_NUMBER: builtins.int
        DETERMINISTIC_FIELD_NUMBER: builtins.int
        SESSION_DEPENDENT_FIELD_NUMBER: builtins.int
        OUTPUT_TYPE_FIELD_NUMBER: builtins.int
        VARIADIC_FIELD_NUMBER: builtins.int
        NORMAL_FIELD_NUMBER: builtins.int
        IMPLEMENTATIONS_FIELD_NUMBER: builtins.int

        @property
        def arguments(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FunctionSignature.Argument]:
            ...

        @property
        def name(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
            ...

        @property
        def description(self) -> global___FunctionSignature.Description:
            ...
        deterministic: builtins.bool
        session_dependent: builtins.bool

        @property
        def output_type(self) -> substrait.ibis.type_expressions_pb2.DerivationExpression:
            ...

        @property
        def variadic(self) -> global___FunctionSignature.FinalArgVariadic:
            ...

        @property
        def normal(self) -> global___FunctionSignature.FinalArgNormal:
            ...

        @property
        def implementations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FunctionSignature.Implementation]:
            ...

        def __init__(self, *, arguments: typing.Optional[typing.Iterable[global___FunctionSignature.Argument]]=..., name: typing.Optional[typing.Iterable[typing.Text]]=..., description: typing.Optional[global___FunctionSignature.Description]=..., deterministic: builtins.bool=..., session_dependent: builtins.bool=..., output_type: typing.Optional[substrait.ibis.type_expressions_pb2.DerivationExpression]=..., variadic: typing.Optional[global___FunctionSignature.FinalArgVariadic]=..., normal: typing.Optional[global___FunctionSignature.FinalArgNormal]=..., implementations: typing.Optional[typing.Iterable[global___FunctionSignature.Implementation]]=...) -> None:
            ...

        def HasField(self, field_name: typing_extensions.Literal['description', b'description', 'final_variable_behavior', b'final_variable_behavior', 'normal', b'normal', 'output_type', b'output_type', 'variadic', b'variadic']) -> builtins.bool:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['arguments', b'arguments', 'description', b'description', 'deterministic', b'deterministic', 'final_variable_behavior', b'final_variable_behavior', 'implementations', b'implementations', 'name', b'name', 'normal', b'normal', 'output_type', b'output_type', 'session_dependent', b'session_dependent', 'variadic', b'variadic']) -> None:
            ...

        def WhichOneof(self, oneof_group: typing_extensions.Literal['final_variable_behavior', b'final_variable_behavior']) -> typing.Optional[typing_extensions.Literal['variadic', 'normal']]:
            ...

    class Aggregate(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        ARGUMENTS_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        DESCRIPTION_FIELD_NUMBER: builtins.int
        DETERMINISTIC_FIELD_NUMBER: builtins.int
        SESSION_DEPENDENT_FIELD_NUMBER: builtins.int
        OUTPUT_TYPE_FIELD_NUMBER: builtins.int
        VARIADIC_FIELD_NUMBER: builtins.int
        NORMAL_FIELD_NUMBER: builtins.int
        ORDERED_FIELD_NUMBER: builtins.int
        MAX_SET_FIELD_NUMBER: builtins.int
        INTERMEDIATE_TYPE_FIELD_NUMBER: builtins.int
        IMPLEMENTATIONS_FIELD_NUMBER: builtins.int

        @property
        def arguments(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FunctionSignature.Argument]:
            ...
        name: typing.Text

        @property
        def description(self) -> global___FunctionSignature.Description:
            ...
        deterministic: builtins.bool
        session_dependent: builtins.bool

        @property
        def output_type(self) -> substrait.ibis.type_expressions_pb2.DerivationExpression:
            ...

        @property
        def variadic(self) -> global___FunctionSignature.FinalArgVariadic:
            ...

        @property
        def normal(self) -> global___FunctionSignature.FinalArgNormal:
            ...
        ordered: builtins.bool
        max_set: builtins.int

        @property
        def intermediate_type(self) -> substrait.ibis.type_pb2.Type:
            ...

        @property
        def implementations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FunctionSignature.Implementation]:
            ...

        def __init__(self, *, arguments: typing.Optional[typing.Iterable[global___FunctionSignature.Argument]]=..., name: typing.Text=..., description: typing.Optional[global___FunctionSignature.Description]=..., deterministic: builtins.bool=..., session_dependent: builtins.bool=..., output_type: typing.Optional[substrait.ibis.type_expressions_pb2.DerivationExpression]=..., variadic: typing.Optional[global___FunctionSignature.FinalArgVariadic]=..., normal: typing.Optional[global___FunctionSignature.FinalArgNormal]=..., ordered: builtins.bool=..., max_set: builtins.int=..., intermediate_type: typing.Optional[substrait.ibis.type_pb2.Type]=..., implementations: typing.Optional[typing.Iterable[global___FunctionSignature.Implementation]]=...) -> None:
            ...

        def HasField(self, field_name: typing_extensions.Literal['description', b'description', 'final_variable_behavior', b'final_variable_behavior', 'intermediate_type', b'intermediate_type', 'normal', b'normal', 'output_type', b'output_type', 'variadic', b'variadic']) -> builtins.bool:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['arguments', b'arguments', 'description', b'description', 'deterministic', b'deterministic', 'final_variable_behavior', b'final_variable_behavior', 'implementations', b'implementations', 'intermediate_type', b'intermediate_type', 'max_set', b'max_set', 'name', b'name', 'normal', b'normal', 'ordered', b'ordered', 'output_type', b'output_type', 'session_dependent', b'session_dependent', 'variadic', b'variadic']) -> None:
            ...

        def WhichOneof(self, oneof_group: typing_extensions.Literal['final_variable_behavior', b'final_variable_behavior']) -> typing.Optional[typing_extensions.Literal['variadic', 'normal']]:
            ...

    class Window(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        class _WindowType:
            ValueType = typing.NewType('ValueType', builtins.int)
            V: typing_extensions.TypeAlias = ValueType

        class _WindowTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[FunctionSignature.Window._WindowType.ValueType], builtins.type):
            DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
            WINDOW_TYPE_UNSPECIFIED: FunctionSignature.Window._WindowType.ValueType
            WINDOW_TYPE_STREAMING: FunctionSignature.Window._WindowType.ValueType
            WINDOW_TYPE_PARTITION: FunctionSignature.Window._WindowType.ValueType

        class WindowType(_WindowType, metaclass=_WindowTypeEnumTypeWrapper):
            pass
        WINDOW_TYPE_UNSPECIFIED: FunctionSignature.Window.WindowType.ValueType
        WINDOW_TYPE_STREAMING: FunctionSignature.Window.WindowType.ValueType
        WINDOW_TYPE_PARTITION: FunctionSignature.Window.WindowType.ValueType
        ARGUMENTS_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        DESCRIPTION_FIELD_NUMBER: builtins.int
        DETERMINISTIC_FIELD_NUMBER: builtins.int
        SESSION_DEPENDENT_FIELD_NUMBER: builtins.int
        INTERMEDIATE_TYPE_FIELD_NUMBER: builtins.int
        OUTPUT_TYPE_FIELD_NUMBER: builtins.int
        VARIADIC_FIELD_NUMBER: builtins.int
        NORMAL_FIELD_NUMBER: builtins.int
        ORDERED_FIELD_NUMBER: builtins.int
        MAX_SET_FIELD_NUMBER: builtins.int
        WINDOW_TYPE_FIELD_NUMBER: builtins.int
        IMPLEMENTATIONS_FIELD_NUMBER: builtins.int

        @property
        def arguments(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FunctionSignature.Argument]:
            ...

        @property
        def name(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
            ...

        @property
        def description(self) -> global___FunctionSignature.Description:
            ...
        deterministic: builtins.bool
        session_dependent: builtins.bool

        @property
        def intermediate_type(self) -> substrait.ibis.type_expressions_pb2.DerivationExpression:
            ...

        @property
        def output_type(self) -> substrait.ibis.type_expressions_pb2.DerivationExpression:
            ...

        @property
        def variadic(self) -> global___FunctionSignature.FinalArgVariadic:
            ...

        @property
        def normal(self) -> global___FunctionSignature.FinalArgNormal:
            ...
        ordered: builtins.bool
        max_set: builtins.int
        window_type: global___FunctionSignature.Window.WindowType.ValueType

        @property
        def implementations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FunctionSignature.Implementation]:
            ...

        def __init__(self, *, arguments: typing.Optional[typing.Iterable[global___FunctionSignature.Argument]]=..., name: typing.Optional[typing.Iterable[typing.Text]]=..., description: typing.Optional[global___FunctionSignature.Description]=..., deterministic: builtins.bool=..., session_dependent: builtins.bool=..., intermediate_type: typing.Optional[substrait.ibis.type_expressions_pb2.DerivationExpression]=..., output_type: typing.Optional[substrait.ibis.type_expressions_pb2.DerivationExpression]=..., variadic: typing.Optional[global___FunctionSignature.FinalArgVariadic]=..., normal: typing.Optional[global___FunctionSignature.FinalArgNormal]=..., ordered: builtins.bool=..., max_set: builtins.int=..., window_type: global___FunctionSignature.Window.WindowType.ValueType=..., implementations: typing.Optional[typing.Iterable[global___FunctionSignature.Implementation]]=...) -> None:
            ...

        def HasField(self, field_name: typing_extensions.Literal['description', b'description', 'final_variable_behavior', b'final_variable_behavior', 'intermediate_type', b'intermediate_type', 'normal', b'normal', 'output_type', b'output_type', 'variadic', b'variadic']) -> builtins.bool:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['arguments', b'arguments', 'description', b'description', 'deterministic', b'deterministic', 'final_variable_behavior', b'final_variable_behavior', 'implementations', b'implementations', 'intermediate_type', b'intermediate_type', 'max_set', b'max_set', 'name', b'name', 'normal', b'normal', 'ordered', b'ordered', 'output_type', b'output_type', 'session_dependent', b'session_dependent', 'variadic', b'variadic', 'window_type', b'window_type']) -> None:
            ...

        def WhichOneof(self, oneof_group: typing_extensions.Literal['final_variable_behavior', b'final_variable_behavior']) -> typing.Optional[typing_extensions.Literal['variadic', 'normal']]:
            ...

    class Description(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        LANGUAGE_FIELD_NUMBER: builtins.int
        BODY_FIELD_NUMBER: builtins.int
        language: typing.Text
        body: typing.Text

        def __init__(self, *, language: typing.Text=..., body: typing.Text=...) -> None:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['body', b'body', 'language', b'language']) -> None:
            ...

    class Implementation(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        class _Type:
            ValueType = typing.NewType('ValueType', builtins.int)
            V: typing_extensions.TypeAlias = ValueType

        class _TypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[FunctionSignature.Implementation._Type.ValueType], builtins.type):
            DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
            TYPE_UNSPECIFIED: FunctionSignature.Implementation._Type.ValueType
            TYPE_WEB_ASSEMBLY: FunctionSignature.Implementation._Type.ValueType
            TYPE_TRINO_JAR: FunctionSignature.Implementation._Type.ValueType

        class Type(_Type, metaclass=_TypeEnumTypeWrapper):
            pass
        TYPE_UNSPECIFIED: FunctionSignature.Implementation.Type.ValueType
        TYPE_WEB_ASSEMBLY: FunctionSignature.Implementation.Type.ValueType
        TYPE_TRINO_JAR: FunctionSignature.Implementation.Type.ValueType
        TYPE_FIELD_NUMBER: builtins.int
        URI_FIELD_NUMBER: builtins.int
        type: global___FunctionSignature.Implementation.Type.ValueType
        uri: typing.Text

        def __init__(self, *, type: global___FunctionSignature.Implementation.Type.ValueType=..., uri: typing.Text=...) -> None:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['type', b'type', 'uri', b'uri']) -> None:
            ...

    class Argument(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        class ValueArgument(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor
            TYPE_FIELD_NUMBER: builtins.int
            CONSTANT_FIELD_NUMBER: builtins.int

            @property
            def type(self) -> substrait.ibis.parameterized_types_pb2.ParameterizedType:
                ...
            constant: builtins.bool

            def __init__(self, *, type: typing.Optional[substrait.ibis.parameterized_types_pb2.ParameterizedType]=..., constant: builtins.bool=...) -> None:
                ...

            def HasField(self, field_name: typing_extensions.Literal['type', b'type']) -> builtins.bool:
                ...

            def ClearField(self, field_name: typing_extensions.Literal['constant', b'constant', 'type', b'type']) -> None:
                ...

        class TypeArgument(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor
            TYPE_FIELD_NUMBER: builtins.int

            @property
            def type(self) -> substrait.ibis.parameterized_types_pb2.ParameterizedType:
                ...

            def __init__(self, *, type: typing.Optional[substrait.ibis.parameterized_types_pb2.ParameterizedType]=...) -> None:
                ...

            def HasField(self, field_name: typing_extensions.Literal['type', b'type']) -> builtins.bool:
                ...

            def ClearField(self, field_name: typing_extensions.Literal['type', b'type']) -> None:
                ...

        class EnumArgument(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor
            OPTIONS_FIELD_NUMBER: builtins.int
            OPTIONAL_FIELD_NUMBER: builtins.int

            @property
            def options(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
                ...
            optional: builtins.bool

            def __init__(self, *, options: typing.Optional[typing.Iterable[typing.Text]]=..., optional: builtins.bool=...) -> None:
                ...

            def ClearField(self, field_name: typing_extensions.Literal['optional', b'optional', 'options', b'options']) -> None:
                ...
        NAME_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        TYPE_FIELD_NUMBER: builtins.int
        ENUM_FIELD_NUMBER: builtins.int
        name: typing.Text

        @property
        def value(self) -> global___FunctionSignature.Argument.ValueArgument:
            ...

        @property
        def type(self) -> global___FunctionSignature.Argument.TypeArgument:
            ...

        @property
        def enum(self) -> global___FunctionSignature.Argument.EnumArgument:
            ...

        def __init__(self, *, name: typing.Text=..., value: typing.Optional[global___FunctionSignature.Argument.ValueArgument]=..., type: typing.Optional[global___FunctionSignature.Argument.TypeArgument]=..., enum: typing.Optional[global___FunctionSignature.Argument.EnumArgument]=...) -> None:
            ...

        def HasField(self, field_name: typing_extensions.Literal['argument_kind', b'argument_kind', 'enum', b'enum', 'type', b'type', 'value', b'value']) -> builtins.bool:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['argument_kind', b'argument_kind', 'enum', b'enum', 'name', b'name', 'type', b'type', 'value', b'value']) -> None:
            ...

        def WhichOneof(self, oneof_group: typing_extensions.Literal['argument_kind', b'argument_kind']) -> typing.Optional[typing_extensions.Literal['value', 'type', 'enum']]:
            ...

    def __init__(self) -> None:
        ...
global___FunctionSignature = FunctionSignature


"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
from ... import substrait
import typing
import typing_extensions
DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class ParameterizedType(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class TypeParameter(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        NAME_FIELD_NUMBER: builtins.int
        BOUNDS_FIELD_NUMBER: builtins.int
        name: typing.Text

        @property
        def bounds(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ParameterizedType]:
            ...

        def __init__(self, *, name: typing.Text=..., bounds: typing.Optional[typing.Iterable[global___ParameterizedType]]=...) -> None:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['bounds', b'bounds', 'name', b'name']) -> None:
            ...

    class IntegerParameter(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        NAME_FIELD_NUMBER: builtins.int
        RANGE_START_INCLUSIVE_FIELD_NUMBER: builtins.int
        RANGE_END_EXCLUSIVE_FIELD_NUMBER: builtins.int
        name: typing.Text

        @property
        def range_start_inclusive(self) -> global___ParameterizedType.NullableInteger:
            ...

        @property
        def range_end_exclusive(self) -> global___ParameterizedType.NullableInteger:
            ...

        def __init__(self, *, name: typing.Text=..., range_start_inclusive: typing.Optional[global___ParameterizedType.NullableInteger]=..., range_end_exclusive: typing.Optional[global___ParameterizedType.NullableInteger]=...) -> None:
            ...

        def HasField(self, field_name: typing_extensions.Literal['range_end_exclusive', b'range_end_exclusive', 'range_start_inclusive', b'range_start_inclusive']) -> builtins.bool:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['name', b'name', 'range_end_exclusive', b'range_end_exclusive', 'range_start_inclusive', b'range_start_inclusive']) -> None:
            ...

    class NullableInteger(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        VALUE_FIELD_NUMBER: builtins.int
        value: builtins.int

        def __init__(self, *, value: builtins.int=...) -> None:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['value', b'value']) -> None:
            ...

    class ParameterizedFixedChar(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        LENGTH_FIELD_NUMBER: builtins.int
        VARIATION_POINTER_FIELD_NUMBER: builtins.int
        NULLABILITY_FIELD_NUMBER: builtins.int

        @property
        def length(self) -> global___ParameterizedType.IntegerOption:
            ...
        variation_pointer: builtins.int
        nullability: substrait.ibis.type_pb2.Type.Nullability.ValueType

        def __init__(self, *, length: typing.Optional[global___ParameterizedType.IntegerOption]=..., variation_pointer: builtins.int=..., nullability: substrait.ibis.type_pb2.Type.Nullability.ValueType=...) -> None:
            ...

        def HasField(self, field_name: typing_extensions.Literal['length', b'length']) -> builtins.bool:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['length', b'length', 'nullability', b'nullability', 'variation_pointer', b'variation_pointer']) -> None:
            ...

    class ParameterizedVarChar(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        LENGTH_FIELD_NUMBER: builtins.int
        VARIATION_POINTER_FIELD_NUMBER: builtins.int
        NULLABILITY_FIELD_NUMBER: builtins.int

        @property
        def length(self) -> global___ParameterizedType.IntegerOption:
            ...
        variation_pointer: builtins.int
        nullability: substrait.ibis.type_pb2.Type.Nullability.ValueType

        def __init__(self, *, length: typing.Optional[global___ParameterizedType.IntegerOption]=..., variation_pointer: builtins.int=..., nullability: substrait.ibis.type_pb2.Type.Nullability.ValueType=...) -> None:
            ...

        def HasField(self, field_name: typing_extensions.Literal['length', b'length']) -> builtins.bool:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['length', b'length', 'nullability', b'nullability', 'variation_pointer', b'variation_pointer']) -> None:
            ...

    class ParameterizedFixedBinary(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        LENGTH_FIELD_NUMBER: builtins.int
        VARIATION_POINTER_FIELD_NUMBER: builtins.int
        NULLABILITY_FIELD_NUMBER: builtins.int

        @property
        def length(self) -> global___ParameterizedType.IntegerOption:
            ...
        variation_pointer: builtins.int
        nullability: substrait.ibis.type_pb2.Type.Nullability.ValueType

        def __init__(self, *, length: typing.Optional[global___ParameterizedType.IntegerOption]=..., variation_pointer: builtins.int=..., nullability: substrait.ibis.type_pb2.Type.Nullability.ValueType=...) -> None:
            ...

        def HasField(self, field_name: typing_extensions.Literal['length', b'length']) -> builtins.bool:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['length', b'length', 'nullability', b'nullability', 'variation_pointer', b'variation_pointer']) -> None:
            ...

    class ParameterizedDecimal(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        SCALE_FIELD_NUMBER: builtins.int
        PRECISION_FIELD_NUMBER: builtins.int
        VARIATION_POINTER_FIELD_NUMBER: builtins.int
        NULLABILITY_FIELD_NUMBER: builtins.int

        @property
        def scale(self) -> global___ParameterizedType.IntegerOption:
            ...

        @property
        def precision(self) -> global___ParameterizedType.IntegerOption:
            ...
        variation_pointer: builtins.int
        nullability: substrait.ibis.type_pb2.Type.Nullability.ValueType

        def __init__(self, *, scale: typing.Optional[global___ParameterizedType.IntegerOption]=..., precision: typing.Optional[global___ParameterizedType.IntegerOption]=..., variation_pointer: builtins.int=..., nullability: substrait.ibis.type_pb2.Type.Nullability.ValueType=...) -> None:
            ...

        def HasField(self, field_name: typing_extensions.Literal['precision', b'precision', 'scale', b'scale']) -> builtins.bool:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['nullability', b'nullability', 'precision', b'precision', 'scale', b'scale', 'variation_pointer', b'variation_pointer']) -> None:
            ...

    class ParameterizedStruct(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        TYPES_FIELD_NUMBER: builtins.int
        VARIATION_POINTER_FIELD_NUMBER: builtins.int
        NULLABILITY_FIELD_NUMBER: builtins.int

        @property
        def types(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ParameterizedType]:
            ...
        variation_pointer: builtins.int
        nullability: substrait.ibis.type_pb2.Type.Nullability.ValueType

        def __init__(self, *, types: typing.Optional[typing.Iterable[global___ParameterizedType]]=..., variation_pointer: builtins.int=..., nullability: substrait.ibis.type_pb2.Type.Nullability.ValueType=...) -> None:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['nullability', b'nullability', 'types', b'types', 'variation_pointer', b'variation_pointer']) -> None:
            ...

    class ParameterizedNamedStruct(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        NAMES_FIELD_NUMBER: builtins.int
        STRUCT_FIELD_NUMBER: builtins.int

        @property
        def names(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
            """list of names in dfs order"""
            pass

        @property
        def struct(self) -> global___ParameterizedType.ParameterizedStruct:
            ...

        def __init__(self, *, names: typing.Optional[typing.Iterable[typing.Text]]=..., struct: typing.Optional[global___ParameterizedType.ParameterizedStruct]=...) -> None:
            ...

        def HasField(self, field_name: typing_extensions.Literal['struct', b'struct']) -> builtins.bool:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['names', b'names', 'struct', b'struct']) -> None:
            ...

    class ParameterizedList(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        TYPE_FIELD_NUMBER: builtins.int
        VARIATION_POINTER_FIELD_NUMBER: builtins.int
        NULLABILITY_FIELD_NUMBER: builtins.int

        @property
        def type(self) -> global___ParameterizedType:
            ...
        variation_pointer: builtins.int
        nullability: substrait.ibis.type_pb2.Type.Nullability.ValueType

        def __init__(self, *, type: typing.Optional[global___ParameterizedType]=..., variation_pointer: builtins.int=..., nullability: substrait.ibis.type_pb2.Type.Nullability.ValueType=...) -> None:
            ...

        def HasField(self, field_name: typing_extensions.Literal['type', b'type']) -> builtins.bool:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['nullability', b'nullability', 'type', b'type', 'variation_pointer', b'variation_pointer']) -> None:
            ...

    class ParameterizedMap(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        VARIATION_POINTER_FIELD_NUMBER: builtins.int
        NULLABILITY_FIELD_NUMBER: builtins.int

        @property
        def key(self) -> global___ParameterizedType:
            ...

        @property
        def value(self) -> global___ParameterizedType:
            ...
        variation_pointer: builtins.int
        nullability: substrait.ibis.type_pb2.Type.Nullability.ValueType

        def __init__(self, *, key: typing.Optional[global___ParameterizedType]=..., value: typing.Optional[global___ParameterizedType]=..., variation_pointer: builtins.int=..., nullability: substrait.ibis.type_pb2.Type.Nullability.ValueType=...) -> None:
            ...

        def HasField(self, field_name: typing_extensions.Literal['key', b'key', 'value', b'value']) -> builtins.bool:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['key', b'key', 'nullability', b'nullability', 'value', b'value', 'variation_pointer', b'variation_pointer']) -> None:
            ...

    class ParameterizedUserDefined(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        TYPE_POINTER_FIELD_NUMBER: builtins.int
        VARIATION_POINTER_FIELD_NUMBER: builtins.int
        NULLABILITY_FIELD_NUMBER: builtins.int
        type_pointer: builtins.int
        variation_pointer: builtins.int
        nullability: substrait.ibis.type_pb2.Type.Nullability.ValueType

        def __init__(self, *, type_pointer: builtins.int=..., variation_pointer: builtins.int=..., nullability: substrait.ibis.type_pb2.Type.Nullability.ValueType=...) -> None:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['nullability', b'nullability', 'type_pointer', b'type_pointer', 'variation_pointer', b'variation_pointer']) -> None:
            ...

    class IntegerOption(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        LITERAL_FIELD_NUMBER: builtins.int
        PARAMETER_FIELD_NUMBER: builtins.int
        literal: builtins.int

        @property
        def parameter(self) -> global___ParameterizedType.IntegerParameter:
            ...

        def __init__(self, *, literal: builtins.int=..., parameter: typing.Optional[global___ParameterizedType.IntegerParameter]=...) -> None:
            ...

        def HasField(self, field_name: typing_extensions.Literal['integer_type', b'integer_type', 'literal', b'literal', 'parameter', b'parameter']) -> builtins.bool:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['integer_type', b'integer_type', 'literal', b'literal', 'parameter', b'parameter']) -> None:
            ...

        def WhichOneof(self, oneof_group: typing_extensions.Literal['integer_type', b'integer_type']) -> typing.Optional[typing_extensions.Literal['literal', 'parameter']]:
            ...
    BOOL_FIELD_NUMBER: builtins.int
    I8_FIELD_NUMBER: builtins.int
    I16_FIELD_NUMBER: builtins.int
    I32_FIELD_NUMBER: builtins.int
    I64_FIELD_NUMBER: builtins.int
    FP32_FIELD_NUMBER: builtins.int
    FP64_FIELD_NUMBER: builtins.int
    STRING_FIELD_NUMBER: builtins.int
    BINARY_FIELD_NUMBER: builtins.int
    TIMESTAMP_FIELD_NUMBER: builtins.int
    DATE_FIELD_NUMBER: builtins.int
    TIME_FIELD_NUMBER: builtins.int
    INTERVAL_YEAR_FIELD_NUMBER: builtins.int
    INTERVAL_DAY_FIELD_NUMBER: builtins.int
    TIMESTAMP_TZ_FIELD_NUMBER: builtins.int
    UUID_FIELD_NUMBER: builtins.int
    FIXED_CHAR_FIELD_NUMBER: builtins.int
    VARCHAR_FIELD_NUMBER: builtins.int
    FIXED_BINARY_FIELD_NUMBER: builtins.int
    DECIMAL_FIELD_NUMBER: builtins.int
    STRUCT_FIELD_NUMBER: builtins.int
    LIST_FIELD_NUMBER: builtins.int
    MAP_FIELD_NUMBER: builtins.int
    USER_DEFINED_FIELD_NUMBER: builtins.int
    USER_DEFINED_POINTER_FIELD_NUMBER: builtins.int
    TYPE_PARAMETER_FIELD_NUMBER: builtins.int

    @property
    def bool(self) -> substrait.ibis.type_pb2.Type.Boolean:
        ...

    @property
    def i8(self) -> substrait.ibis.type_pb2.Type.I8:
        ...

    @property
    def i16(self) -> substrait.ibis.type_pb2.Type.I16:
        ...

    @property
    def i32(self) -> substrait.ibis.type_pb2.Type.I32:
        ...

    @property
    def i64(self) -> substrait.ibis.type_pb2.Type.I64:
        ...

    @property
    def fp32(self) -> substrait.ibis.type_pb2.Type.FP32:
        ...

    @property
    def fp64(self) -> substrait.ibis.type_pb2.Type.FP64:
        ...

    @property
    def string(self) -> substrait.ibis.type_pb2.Type.String:
        ...

    @property
    def binary(self) -> substrait.ibis.type_pb2.Type.Binary:
        ...

    @property
    def timestamp(self) -> substrait.ibis.type_pb2.Type.Timestamp:
        ...

    @property
    def date(self) -> substrait.ibis.type_pb2.Type.Date:
        ...

    @property
    def time(self) -> substrait.ibis.type_pb2.Type.Time:
        ...

    @property
    def interval_year(self) -> substrait.ibis.type_pb2.Type.IntervalYear:
        ...

    @property
    def interval_day(self) -> substrait.ibis.type_pb2.Type.IntervalDay:
        ...

    @property
    def timestamp_tz(self) -> substrait.ibis.type_pb2.Type.TimestampTZ:
        ...

    @property
    def uuid(self) -> substrait.ibis.type_pb2.Type.UUID:
        ...

    @property
    def fixed_char(self) -> global___ParameterizedType.ParameterizedFixedChar:
        ...

    @property
    def varchar(self) -> global___ParameterizedType.ParameterizedVarChar:
        ...

    @property
    def fixed_binary(self) -> global___ParameterizedType.ParameterizedFixedBinary:
        ...

    @property
    def decimal(self) -> global___ParameterizedType.ParameterizedDecimal:
        ...

    @property
    def struct(self) -> global___ParameterizedType.ParameterizedStruct:
        ...

    @property
    def list(self) -> global___ParameterizedType.ParameterizedList:
        ...

    @property
    def map(self) -> global___ParameterizedType.ParameterizedMap:
        ...

    @property
    def user_defined(self) -> global___ParameterizedType.ParameterizedUserDefined:
        ...
    user_defined_pointer: builtins.int
    'Deprecated in favor of user_defined, which allows nullability and\n    variations to be specified. If user_defined_pointer is encountered,\n    treat it as being non-nullable and having the default variation.\n    '

    @property
    def type_parameter(self) -> global___ParameterizedType.TypeParameter:
        ...

    def __init__(self, *, bool: typing.Optional[substrait.ibis.type_pb2.Type.Boolean]=..., i8: typing.Optional[substrait.ibis.type_pb2.Type.I8]=..., i16: typing.Optional[substrait.ibis.type_pb2.Type.I16]=..., i32: typing.Optional[substrait.ibis.type_pb2.Type.I32]=..., i64: typing.Optional[substrait.ibis.type_pb2.Type.I64]=..., fp32: typing.Optional[substrait.ibis.type_pb2.Type.FP32]=..., fp64: typing.Optional[substrait.ibis.type_pb2.Type.FP64]=..., string: typing.Optional[substrait.ibis.type_pb2.Type.String]=..., binary: typing.Optional[substrait.ibis.type_pb2.Type.Binary]=..., timestamp: typing.Optional[substrait.ibis.type_pb2.Type.Timestamp]=..., date: typing.Optional[substrait.ibis.type_pb2.Type.Date]=..., time: typing.Optional[substrait.ibis.type_pb2.Type.Time]=..., interval_year: typing.Optional[substrait.ibis.type_pb2.Type.IntervalYear]=..., interval_day: typing.Optional[substrait.ibis.type_pb2.Type.IntervalDay]=..., timestamp_tz: typing.Optional[substrait.ibis.type_pb2.Type.TimestampTZ]=..., uuid: typing.Optional[substrait.ibis.type_pb2.Type.UUID]=..., fixed_char: typing.Optional[global___ParameterizedType.ParameterizedFixedChar]=..., varchar: typing.Optional[global___ParameterizedType.ParameterizedVarChar]=..., fixed_binary: typing.Optional[global___ParameterizedType.ParameterizedFixedBinary]=..., decimal: typing.Optional[global___ParameterizedType.ParameterizedDecimal]=..., struct: typing.Optional[global___ParameterizedType.ParameterizedStruct]=..., list: typing.Optional[global___ParameterizedType.ParameterizedList]=..., map: typing.Optional[global___ParameterizedType.ParameterizedMap]=..., user_defined: typing.Optional[global___ParameterizedType.ParameterizedUserDefined]=..., user_defined_pointer: builtins.int=..., type_parameter: typing.Optional[global___ParameterizedType.TypeParameter]=...) -> None:
        ...

    def HasField(self, field_name: typing_extensions.Literal['binary', b'binary', 'bool', b'bool', 'date', b'date', 'decimal', b'decimal', 'fixed_binary', b'fixed_binary', 'fixed_char', b'fixed_char', 'fp32', b'fp32', 'fp64', b'fp64', 'i16', b'i16', 'i32', b'i32', 'i64', b'i64', 'i8', b'i8', 'interval_day', b'interval_day', 'interval_year', b'interval_year', 'kind', b'kind', 'list', b'list', 'map', b'map', 'string', b'string', 'struct', b'struct', 'time', b'time', 'timestamp', b'timestamp', 'timestamp_tz', b'timestamp_tz', 'type_parameter', b'type_parameter', 'user_defined', b'user_defined', 'user_defined_pointer', b'user_defined_pointer', 'uuid', b'uuid', 'varchar', b'varchar']) -> builtins.bool:
        ...

    def ClearField(self, field_name: typing_extensions.Literal['binary', b'binary', 'bool', b'bool', 'date', b'date', 'decimal', b'decimal', 'fixed_binary', b'fixed_binary', 'fixed_char', b'fixed_char', 'fp32', b'fp32', 'fp64', b'fp64', 'i16', b'i16', 'i32', b'i32', 'i64', b'i64', 'i8', b'i8', 'interval_day', b'interval_day', 'interval_year', b'interval_year', 'kind', b'kind', 'list', b'list', 'map', b'map', 'string', b'string', 'struct', b'struct', 'time', b'time', 'timestamp', b'timestamp', 'timestamp_tz', b'timestamp_tz', 'type_parameter', b'type_parameter', 'user_defined', b'user_defined', 'user_defined_pointer', b'user_defined_pointer', 'uuid', b'uuid', 'varchar', b'varchar']) -> None:
        ...

    def WhichOneof(self, oneof_group: typing_extensions.Literal['kind', b'kind']) -> typing.Optional[typing_extensions.Literal['bool', 'i8', 'i16', 'i32', 'i64', 'fp32', 'fp64', 'string', 'binary', 'timestamp', 'date', 'time', 'interval_year', 'interval_day', 'timestamp_tz', 'uuid', 'fixed_char', 'varchar', 'fixed_binary', 'decimal', 'struct', 'list', 'map', 'user_defined', 'user_defined_pointer', 'type_parameter']]:
        ...
global___ParameterizedType = ParameterizedType


"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.any_pb2
import google.protobuf.descriptor
import google.protobuf.message
import typing
import typing_extensions
DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class SimpleExtensionURI(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    EXTENSION_URI_ANCHOR_FIELD_NUMBER: builtins.int
    URI_FIELD_NUMBER: builtins.int
    extension_uri_anchor: builtins.int
    'A surrogate key used in the context of a single plan used to reference the\n    URI associated with an extension.\n    '
    uri: typing.Text
    'The URI where this extension YAML can be retrieved. This is the "namespace"\n    of this extension.\n    '

    def __init__(self, *, extension_uri_anchor: builtins.int=..., uri: typing.Text=...) -> None:
        ...

    def ClearField(self, field_name: typing_extensions.Literal['extension_uri_anchor', b'extension_uri_anchor', 'uri', b'uri']) -> None:
        ...
global___SimpleExtensionURI = SimpleExtensionURI

class SimpleExtensionDeclaration(google.protobuf.message.Message):
    """Describes a mapping between a specific extension entity and the uri where
    that extension can be found.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class ExtensionType(google.protobuf.message.Message):
        """Describes a Type"""
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        EXTENSION_URI_REFERENCE_FIELD_NUMBER: builtins.int
        TYPE_ANCHOR_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        extension_uri_reference: builtins.int
        'references the extension_uri_anchor defined for a specific extension URI.'
        type_anchor: builtins.int
        'A surrogate key used in the context of a single plan to reference a\n        specific extension type\n        '
        name: typing.Text
        'the name of the type in the defined extension YAML.'

        def __init__(self, *, extension_uri_reference: builtins.int=..., type_anchor: builtins.int=..., name: typing.Text=...) -> None:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['extension_uri_reference', b'extension_uri_reference', 'name', b'name', 'type_anchor', b'type_anchor']) -> None:
            ...

    class ExtensionTypeVariation(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        EXTENSION_URI_REFERENCE_FIELD_NUMBER: builtins.int
        TYPE_VARIATION_ANCHOR_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        extension_uri_reference: builtins.int
        'references the extension_uri_anchor defined for a specific extension URI.'
        type_variation_anchor: builtins.int
        'A surrogate key used in the context of a single plan to reference a\n        specific type variation\n        '
        name: typing.Text
        'the name of the type in the defined extension YAML.'

        def __init__(self, *, extension_uri_reference: builtins.int=..., type_variation_anchor: builtins.int=..., name: typing.Text=...) -> None:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['extension_uri_reference', b'extension_uri_reference', 'name', b'name', 'type_variation_anchor', b'type_variation_anchor']) -> None:
            ...

    class ExtensionFunction(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        EXTENSION_URI_REFERENCE_FIELD_NUMBER: builtins.int
        FUNCTION_ANCHOR_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        extension_uri_reference: builtins.int
        'references the extension_uri_anchor defined for a specific extension URI.'
        function_anchor: builtins.int
        'A surrogate key used in the context of a single plan to reference a\n        specific function\n        '
        name: typing.Text
        'A simple name if there is only one impl for the function within the YAML.\n        A compound name, referencing that includes type short names if there is\n        more than one impl per name in the YAML.\n        '

        def __init__(self, *, extension_uri_reference: builtins.int=..., function_anchor: builtins.int=..., name: typing.Text=...) -> None:
            ...

        def ClearField(self, field_name: typing_extensions.Literal['extension_uri_reference', b'extension_uri_reference', 'function_anchor', b'function_anchor', 'name', b'name']) -> None:
            ...
    EXTENSION_TYPE_FIELD_NUMBER: builtins.int
    EXTENSION_TYPE_VARIATION_FIELD_NUMBER: builtins.int
    EXTENSION_FUNCTION_FIELD_NUMBER: builtins.int

    @property
    def extension_type(self) -> global___SimpleExtensionDeclaration.ExtensionType:
        ...

    @property
    def extension_type_variation(self) -> global___SimpleExtensionDeclaration.ExtensionTypeVariation:
        ...

    @property
    def extension_function(self) -> global___SimpleExtensionDeclaration.ExtensionFunction:
        ...

    def __init__(self, *, extension_type: typing.Optional[global___SimpleExtensionDeclaration.ExtensionType]=..., extension_type_variation: typing.Optional[global___SimpleExtensionDeclaration.ExtensionTypeVariation]=..., extension_function: typing.Optional[global___SimpleExtensionDeclaration.ExtensionFunction]=...) -> None:
        ...

    def HasField(self, field_name: typing_extensions.Literal['extension_function', b'extension_function', 'extension_type', b'extension_type', 'extension_type_variation', b'extension_type_variation', 'mapping_type', b'mapping_type']) -> builtins.bool:
        ...

    def ClearField(self, field_name: typing_extensions.Literal['extension_function', b'extension_function', 'extension_type', b'extension_type', 'extension_type_variation', b'extension_type_variation', 'mapping_type', b'mapping_type']) -> None:
        ...

    def WhichOneof(self, oneof_group: typing_extensions.Literal['mapping_type', b'mapping_type']) -> typing.Optional[typing_extensions.Literal['extension_type', 'extension_type_variation', 'extension_function']]:
        ...
global___SimpleExtensionDeclaration = SimpleExtensionDeclaration

class AdvancedExtension(google.protobuf.message.Message):
    """A generic object that can be used to embed additional extension information
    into the serialized substrait plan.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    OPTIMIZATION_FIELD_NUMBER: builtins.int
    ENHANCEMENT_FIELD_NUMBER: builtins.int

    @property
    def optimization(self) -> google.protobuf.any_pb2.Any:
        """An optimization is helpful information that don't influence semantics. May
        be ignored by a consumer.
        """
        pass

    @property
    def enhancement(self) -> google.protobuf.any_pb2.Any:
        """An enhancement alter semantics. Cannot be ignored by a consumer."""
        pass

    def __init__(self, *, optimization: typing.Optional[google.protobuf.any_pb2.Any]=..., enhancement: typing.Optional[google.protobuf.any_pb2.Any]=...) -> None:
        ...

    def HasField(self, field_name: typing_extensions.Literal['enhancement', b'enhancement', 'optimization', b'optimization']) -> builtins.bool:
        ...

    def ClearField(self, field_name: typing_extensions.Literal['enhancement', b'enhancement', 'optimization', b'optimization']) -> None:
        ...
global___AdvancedExtension = AdvancedExtension